@@ -20,12 +20,14 @@ bool parse_cycle_direction(char *s, cycle_dir_t *d);
 bool parse_circulate_direction(char *s, circulate_dir_t *d);
 bool parse_history_direction(char *s, history_dir_t *d);
 bool parse_flip(char *s, flip_t *f);
+bool parse_resize_handle(char *s, resize_handle_t *h);
+bool parse_modifier_mask(char *s, uint16_t *m);
 bool parse_pointer_action(char *s, pointer_action_t *a);
 bool parse_child_polarity(char *s, child_polarity_t *p);
 bool parse_degree(char *s, int *d);
-bool parse_id(char *s, uint32_t *i);
+bool parse_id(char *s, uint32_t *id);
 bool parse_bool_declaration(char *s, char **key, bool *value, alter_state_t *state);
-bool parse_index(char *s, int *i);
+bool parse_index(char *s, uint16_t *idx);
 bool parse_rectangle(char *s, xcb_rectangle_t *r);
 bool parse_subscriber_mask(char *s, subscriber_mask_t *mask);
 bool parse_monitor_modifiers(char *desc, monitor_select_t *sel);
